/* -*- C++ -*-
 * Copyright ©2008 Hugo Mills <hugo@carfax.org.uk>
 *
 * This software is distributed under the terms of the GNU GPL v2
 * For more information on the GPL, see the file COPYING or 
 * visit http://www.gnu.org/
 *
 * This software is distributed without warranty
 */

#include "content_terminator.h"

#include "contentgroup.h"
#include "output_canvas.h"
#include "xfview_translate.h"

#include "magellan/parsertools.h"
#include "magellan/options.h"
#include "plugins.h"

extern "C" PluginInfo* getmodules(int i)
{
	if(i==0)
	{
		PluginInfo_Content* pi = new PluginInfo_Content();

		pi->name = "Terminator line";
		pi->author = "Hugo Mills";
		pi->copyright = "©2008 Hugo Mills";
		pi->version = 0x00900;
		pi->parser = Content_Terminator::parse;
		return pi;
	}
	else
		return NULL;
}

Content_Terminator::Content_Terminator()
	: day(NULL), night(NULL), sunpos(NULL),
	  blend(0), bwidth(0.0)
{
}

Content_Terminator::~Content_Terminator()
{
}

Content* Content_Terminator::parse(
	ConfigLexer* lex,
	const std::string& param_type,
	const std::string& param_subtype
	)
{
	if(param_type != "terminator")
		return NULL;

	Content_Terminator* ct = new Content_Terminator();

	std::string blend_keyword = "linear";

	while(!lex->eos()) {
		// We have some top-level config options
		if(readkeyedvalue(lex, "boundary-width", &ct->bwidth)
		   || readkeyedvalue(lex, "boundary-blend", &blend_keyword)
			)
		{
			continue;
		}

		// The remaining options are complex
		if(lex->type() != TOKEN)
		{
			// Bitch about it
			std::cerr << "Type is not a token (line "
					  << lex->lineno() << ")" << std::endl;
			return NULL;
		}

		std::string option = lex->value();
		lex->next();
		
		if(lex->type() != CHAR || lex->value() != "{")
		{
			// Bitch about it
			std::cerr << "Expected { but not found (line "
					  << lex->lineno() << ")" << std::endl;
			return NULL;
		}
		lex->next();

		// Sun definition. This is an orbit.
		if(option == "sun")
		{
			std::string type = "";
			std::string subtype = "";
			readkeyedvalue(lex, "type", &type); // Read it or not, we
												// don't care too much
												// here...

			try {
				XfOrbit* res = readplugin<XfOrbit>(
					lex, type, subtype,
					orbit_plugins.begin(),
					orbit_plugins.end()
					);
				if(res)
					ct->sunpos = res;
			} catch(...) {
				std::cerr << "Caught error in sun definition" << std::endl
						  << " at line " << lex->lineno() 
						  << " of config file" << std::endl;
			}

			continue;
		}

		// Day and night are both content groups
		if(option == "day")
		{
			// Read a content group
			ct->day = ContentGroup::parse(lex, "", "");
			if(opts->debug_parse() >= 4 || opts->debug_render() >= 2)
				std::cerr << "      terminator-day created at " << ct->day << std::endl;
			lex->next();
			continue;
		}

		if(option == "night")
		{
			ct->night = ContentGroup::parse(lex, "", "");
			if(opts->debug_parse() >= 4 || opts->debug_render() >= 2)
				std::cerr << "      terminator-night created at " << ct->day << std::endl;
			lex->next();
			continue;
		}

		// Unhandled situation. Throw an error
		throw(UnknownConfigToken(lex));
	}

	if(ct->day == NULL)
		throw(PluginInitError("Required section \"day\" missing."));
	if(ct->night == NULL)
		throw(PluginInitError("Required section \"night\" missing."));
	if(ct->sunpos == NULL)
		throw(PluginInitError("Required section \"sun\" missing."));

	return ct;
}

void Content_Terminator::plot(
	Output* bitmap,
	XfView* viewport,
	XfMap* projection,
	XfOrbit* orbit,
	XfSphere* sphere
	)
{
	// Create temporary canvases to hold day and night
	Output_Canvas* day_canvas = new Output_Canvas(bitmap->width(), bitmap->height());
	Output_Canvas* night_canvas = new Output_Canvas(bitmap->width(), bitmap->height());

	// Translate the supplied viewport into something that has an
	// origin at (0,0), to match our canvases
	XfView* vp = new XfView_Translate(viewport,
									  viewport->left(),
									  viewport->bottom());
	// Plot day and night separately
	day->plot(day_canvas, viewport, projection, orbit, sphere);
	night->plot(night_canvas, vp, projection, orbit, sphere);

	ivec2 bm_pos;
	dvec2 map_pos;
	dvec3 shifted_pos;
	dvec3 globe_pos;
	dvec2 latlong;

/*
	map_pos[0] = map_pos[1] = 0.0;
	vp->f(bm_pos, map_pos);
	std::cerr << "Bitmap pos at map centre = " << bm_pos[0] << ", " << bm_pos[1] << std::endl;
	viewport->f(bm_pos, map_pos);
	std::cerr << "Bitmap pos at map centre = " << bm_pos[0] << ", " << bm_pos[1] << std::endl;
*/

/*
	for(bm_pos[0] = 0; bm_pos[0] < day_canvas->width(); bm_pos[0]++)
		for(bm_pos[1] = 0; bm_pos[1] < day_canvas->height(); bm_pos[1]++)
		{
			night_canvas->setpixel(bm_pos, 253, 0, 0);
		}
*/

	ivec3 colour;

	// Work out the direction of the sun from the centre of the globe
	dvec3 sun;
	shifted_pos[0] = 0.0;
	shifted_pos[1] = 0.0;
	shifted_pos[2] = 1.0;
	sunpos->g(sun, shifted_pos);

	// For each pixel on those canvases, compute the relative sun
	// position and blend
	for(bm_pos[1] = viewport->bottom();
		bm_pos[1] < viewport->top();
		bm_pos[1]++)
	{
		for(bm_pos[0] = viewport->left();
			bm_pos[0] < viewport->right();
			bm_pos[0]++)
		{
			if(!viewport->g(map_pos, bm_pos))
				continue;

			// Compute the position of this point on the globe's surface
			if(projection->g(shifted_pos, map_pos)
			   && orbit->g(globe_pos, shifted_pos))
			{
				// globe_pos is now the (x,y,z) position on the
				// earth's surface of the point in question.

				// Compare to the sun position: dot product gives
				// cosine of the angle between the two unit vectors
				double cossigma = 0.0;
				cossigma += globe_pos[0]*sun[0];
				cossigma += globe_pos[1]*sun[1];
				cossigma += globe_pos[2]*sun[2];

				if(cossigma < 0)
					night_canvas->getpixel(colour, bm_pos);
				else
					day_canvas->getpixel(colour, bm_pos);

				bitmap->setpixel(bm_pos, colour[0], colour[1], colour[2]);
			}
		}
	}

	// Tidy up
	delete night_canvas;
	delete day_canvas;
}

bool Content_Terminator::isbound(void) const
{
	return day->isbound() && night->isbound() && sunpos->isbound();
}

Plugin* Content_Terminator::bind(Configuration* conf)
{
	if(!day->isbound())
	{
		Plugin* p = day->bind(conf);
		if(p == NULL)
			return NULL;
		day = dynamic_cast<Content*>(p);
	}
	if(!night->isbound())
	{
		Plugin* p = night->bind(conf);
		if(p == NULL)
			return NULL;
		night = dynamic_cast<Content*>(p);
	}
	if(!sunpos->isbound())
	{
		Plugin* p = sunpos->bind(conf);
		if(p == NULL)
			return NULL;
		sunpos = dynamic_cast<XfOrbit*>(p);
	}

	return this;
}
